/* Modified from glibc: io/tst-copy_file_range.c for oscomp.

   Tests for splice.
   Copyright (C) 2017-2023 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, see
   <https://www.gnu.org/licenses/>.  */

#define _GNU_SOURCE

#include <fcntl.h>
#include <inttypes.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>

/* array_length (VAR) is the number of elements in the array VAR.  VAR
   must evaluate to an array, not a pointer.  */
#define array_length(var)                                                   \
  (sizeof(var) / sizeof((var)[0]) +                                         \
   0 * sizeof(struct {                                                      \
     _Static_assert(                                                        \
         !__builtin_types_compatible_p(__typeof(var), __typeof(&(var)[0])), \
         "argument must be an array");                                      \
   }))

/* array_end (VAR) is a pointer one past the end of the array VAR.
   VAR must evaluate to an array, not a pointer.  */
#define array_end(var) (&(var)[array_length(var)])

#define max(a, b) ((a) > (b) ? (a) : (b))

int to_kill = 0;
/* The size chosen by the test harness.  */
static int current_size;

void fail_exit(const char* format, ...) {
  if (to_kill) {
    kill(to_kill, SIGTERM);
  }

  va_list ap;
  va_start(ap, format);
  vfprintf(stderr, format, ap);
  va_end(ap);
  exit(EXIT_FAILURE);
}

#define die(fmt_, ...) \
  fail_exit("%s:%d: " fmt_ "\n", __FILE__, __LINE__, ##__VA_ARGS__)

#define TEST_COMPARE(a_, b_) \
  {                          \
    int a__ = (a_);\
    int b__ = (b_);\
    if (a__ != b__) {          \
      die("compare failed a = %d, b = %d", a__, b__); \
    }                        \
  }

#define TEST_VERIFY(a_)     \
  {                         \
    if (!(a_)) {              \
      die("verify failed"); \
    }                       \
  }

void xwrite(int fd, const void* buffer, size_t length) {
  const char* p = buffer;
  const char* end = p + length;
  while (p < end) {
    ssize_t ret = write(fd, p, end - p);
    if (ret < 0)
      die("write of %zu bytes failed after %td", length,
          p - (const char*)buffer);
    if (ret == 0)
      die("write return 0 after writing %td bytes of %zu",
          p - (const char*)buffer, length);
    p += ret;
  }
}

void xpipe(int *pipe_des) {
  int ret = pipe(pipe_des);
  if (ret < 0) {
    die("create pipe failed!");
  }
}

long long xlseek(int fd, long long offset, int whence) {
  long long result = lseek(fd, offset, whence);
  if (result < 0)
    die("lseek(%d, %lld, %d)", fd, offset, whence);
  return result;
}

void* xmalloc(size_t n) {
  void* p;
  p = malloc(n);
  if (p == NULL)
    die("malloc %zu", n);
  return p;
}

int xopen(const char* path, int flags, mode_t mode) {
  int ret = open(path, flags, mode);
  if (ret < 0)
    die("open (\"%s\", 0x%x, 0%o)", path, flags, mode);
  return ret;
}

void xfstat(int fd, struct stat* result) {
  if (fstat(fd, result) != 0)
    die("fstat (%d)", fd);
}

void xftruncate(int fd, long long length) {
  if (ftruncate(fd, length) != 0)
    die("ftruncate (%d, %lld)", fd, length);
}

ssize_t xsplice(int fdin, loff_t *offin, int fdout,
  loff_t *offout, size_t len, unsigned int flags
) {
  int acc = 0;
  int cur = 0;
  while ((cur = splice(fdin, offin, fdout, offout, len - acc, flags)) != 0) {
    if (cur < 0) {
      die("splice returns negative number");
    }
    acc += cur;
  }
  return acc;
}

void xread(int fd, void *buf, size_t nbytes) {
  int acc = 0;
  int cur = 0;
  while ((cur = read(fd, buf + acc, nbytes - acc)) != 0) {
    if (cur < 0) {
      die("read failed! readed bytes: %d", acc);
    }
    acc += cur;
  }
  if (acc != nbytes) {
    die("read return 0. readed bytes: %d", acc);
  }
}


#define xclose close
#define test_verbose 0

#ifdef O_LARGEFILE
#undef O_LARGEFILE
#endif
#define O_LARGEFILE 0

int create_temp_file(const char* prefix, char** name) {
  *name = xmalloc(strlen(prefix) + 8);
  strcpy(*name, prefix);
  strcat(*name, "XXXXXX");
  int fd = mkstemp(*name);
  if (fd < 0)
    die("mkstemp: %d", fd);
  // printf("created: %s\n", *name);
  return fd;
}


/* Name and descriptors of the input files.  Files are truncated and
   reopened (with O_RDWR) between tests.  */
static char* infile;
static int infd;
static char* outfile;
static int outfd;

/* Input and output offsets.  Set according to do_inoff and do_outoff
   before the test.  The offsets themselves are always set to
   zero.  */
static off_t inoff;
static off_t outoff;

/* These are a collection of copy sizes used in tests.    */
enum { maximum_size = 99999 };
static const int typical_sizes[] = {0,    1,    2,    3,    1024,        2048,
                                    4096, 8191, 8192, 8193, maximum_size};
static const int min_filesize = 16;


/* The random contents of this array can be used as a pattern to check
   for correct write operations.  */
static unsigned char random_data[maximum_size];
static unsigned char zeros[maximum_size] = {0};



// pipe's read/write num of bytes is less than len
int pipe_less_data = 0;

/* using current_size, copy file to pipe.  */
static void file_to_pipe_test(void) {
  int fds[2]; // 0 for read, 1 for write
  int child;

  if (pipe_less_data != 0 && current_size <= pipe_less_data) {
    return;
  }

  xpipe(fds);
  to_kill = getpid();

  if ((child = fork()) == 0) {
    // copy from file to pipe
    close(fds[0]);

    outfd = fds[1];
    inoff = 5;
    int infd_seek = 1;

    infd = create_temp_file("tst-splice-in-", &infile);
    xwrite(infd, zeros, inoff);
    xwrite(infd, random_data, max(current_size - pipe_less_data, 0));

    xlseek(infd, infd_seek, SEEK_SET);
    TEST_COMPARE(xsplice(infd, &inoff, outfd, NULL, current_size, 0), max(current_size - pipe_less_data, 0));

    // check inoff
    TEST_COMPARE(inoff, 5 + max(current_size - pipe_less_data, 0));
    // check file offset
    TEST_COMPARE(xlseek(infd, 0, SEEK_CUR), infd_seek);

    close(infd);
    close(fds[1]);
    remove(infile);
    exit(0);
  } else {
    to_kill = child;

    // read
    close(fds[1]);
    
    int read_fd = fds[0];
    char *bytes = xmalloc(max(current_size, min_filesize));
    // when pipe_less_data != 0, consume less data!
    xread(read_fd, bytes, max(current_size - pipe_less_data, 0));
    close(read_fd);
    // compare whether the copied data is consistent
    TEST_VERIFY(memcmp(random_data, bytes, max(current_size - pipe_less_data, 0)) == 0);
    free(bytes);

    close(fds[0]);
  }
}

/* using current_size, copy pipe to file.  */
static void pipe_to_file_test(void) {
  int fds[2]; // 0 for read, 1 for write
  xpipe(fds);

  if (fork() == 0) {
    // write
    close(fds[0]);
    int write_fd = fds[1];

    if (pipe_less_data) {
      xwrite(write_fd, random_data, max(current_size - pipe_less_data, 0));
    } else {
      xwrite(write_fd, random_data, current_size);
    }
    close(write_fd);

    exit(0);
  } else {
    close(fds[1]);
    // copy from pipe to file
    int length;
    int outfd_seek;
    
    length = current_size;
    infd = fds[0];
    outfd = create_temp_file("tst-splice-out-", &outfile);
    xwrite(outfd, zeros, min_filesize);

    outfd_seek = 3;

    xlseek(outfd, outfd_seek, SEEK_SET);
    outoff = 2;
    if (pipe_less_data) {
      TEST_COMPARE(xsplice(infd, NULL, outfd, &outoff, length, 0), max(length - pipe_less_data, 0));
    } else {
      TEST_COMPARE(xsplice(infd, NULL, outfd, &outoff, length, 0), length);
    }

    // check outoff
    TEST_COMPARE(outoff, 2 + max(length - pipe_less_data, 0));
    // check file offset
    TEST_COMPARE(xlseek(outfd, 0, SEEK_CUR), outfd_seek);
    struct stat st;
    xfstat(outfd, &st);
    TEST_COMPARE(st.st_size, max(outoff, min_filesize));

    // Compare whether the copied data is consistent
    xlseek(outfd, 0, SEEK_SET);
    char *bytes = xmalloc(st.st_size);
    TEST_COMPARE(read(outfd, bytes, st.st_size), st.st_size);
    TEST_VERIFY(memcmp(bytes + 2, random_data, max(length - pipe_less_data, 0)) == 0);
    free(bytes);

    close(fds[0]);
    close(outfd);
    remove(outfile);
  }
}

void corner_case() {
  if (current_size == 0) {
    return;
  }

  int fds[2]; // 0 for read, 1 for write
  xpipe(fds);
  int filefd = create_temp_file("tst-splice-in-", &infile);
  xwrite(filefd, zeros, 5);
  inoff = 8;
  TEST_COMPARE(xsplice(filefd, &inoff, fds[1], NULL, current_size, 0), 0);
  inoff = -1;
  TEST_VERIFY(splice(filefd, &inoff, fds[1], NULL, current_size, 0) < 0);
  outoff = -5;
  TEST_VERIFY(splice(fds[0], NULL, filefd, &outoff, current_size, 0) < 0);
  remove(infile);
}

void do_test() {
  // make random data
  for (unsigned char* p = random_data; p < array_end(random_data); ++p)
    *p = rand() >> 24;

  // use multiple sizes
  for (const int *size = typical_sizes; size < array_end(typical_sizes); size++) {
    current_size = *size;

    #if CASE == 1
      pipe_to_file_test();
    #elif CASE == 2
      file_to_pipe_test();
    #elif CASE == 3
      pipe_less_data = 5;
      pipe_to_file_test();
    #elif CASE == 4
      pipe_less_data = 3;
      file_to_pipe_test();
    #else
      corner_case();
    #endif

    printf("Pass the test with size=%d\n", current_size);
  }
}

int main(void) {
  do_test();
  printf("copy-file-range-test: passed case %d\n", CASE);
  return 0;
}